/********************************************************************\

  General code to start a server of the Evidence Control System
  
  - The server is started with the given name.
  - DIM exit and error handlers are implemented.
  - The Status service is published (special format, see below).
    It can be updated with the State() method. The text will also be logged.
  - If the severity of a State() call is FATAL, exit() will be called (with
    this severity, the call to State() is guranteed not to return).
  - Configuration data can be requested by GetConfig(). 
  - Signal handlers to ignore common signals are installed.
    These signals will then cause pause() to return which can be used
	by the application to terminate gracefully.
  - The static method ToString() converts the contents of a
    DIMInfo service into text
	
  All memory allocated by the non-static methods will be freed by the
  class destructor.
  
  Oliver Grimm, December 2009
 
\********************************************************************/

#include "Evidence.h"

using namespace std;

//////////////////////////
// EvidenceServer Class //
//////////////////////////

EvidenceServer *ThisServer;

// Constructor starts server with given name
EvidenceServer::EvidenceServer(const char *Name) {

  // Initialize
  Status = NULL;
  ExitRequest = false;
  ThisServer = this;
  
  // Catch some signals
  signal(SIGQUIT, &SignalHandler);  // CTRL-Backspace
  signal(SIGTERM, &SignalHandler);  // Termination signal
  signal(SIGINT, &SignalHandler);   // CTRL-C
  signal(SIGHUP, &SignalHandler);   // Terminal closed
  
  // Catch C++ unhandled exceptions
  set_terminate(Terminate);
  
  // Start server
  static char Init[] = "Server started";
  Status = new DimService((string(Name) + "/Status").c_str(), (char *) "C", Init, sizeof(Init));

  start(Name);
  addExitHandler(this);
}

// Destructor: Free memory
EvidenceServer::~EvidenceServer() {

  State(INFO, "Server stopped");
  
  for (unsigned int i=0; i<ConfigList.size(); i++) {
  	delete[] ConfigList[i].Name;
	delete[] ConfigList[i].Value;
  }
}

// DIM exit handler
void EvidenceServer::exitHandler(int Code) {
  State(INFO, "Exit handler called (DIM exit code %d)", Code);
  exit(EXIT_SUCCESS);
}

// DIM error handler
void EvidenceServer::errorHandler(int Severity, int Code, char *Message) {   
  State(ERROR, "%s (DIM error code %d, severity %d)\n", Message, Code, Severity);
}

// Set status of server
//
// The message format is special: after the string-terminating '\0' the Severity 
// is given, terminated by another '\0'  The buffer for the DIM service must have
// the same lifetime as the DIM service. If Severity is FATAL, exit() will be invoked.
void EvidenceServer::State(StateType Severity, const char *Format, ...) {

  static const char* StateString[] = {"Info", "Warn", "Error", "Fatal"};
  static char ErrorString[] = "vasprintf() failed in State()";
  static char SBuf[STATUS_SIZE];
  char TBuf[STATUS_SIZE];
  char *Tmp;
  
  // Assemble message from application
  va_list ArgumentPointer;
  va_start(ArgumentPointer, Format);
  if (vasprintf(&Tmp, Format, ArgumentPointer) == -1) Tmp = ErrorString;
  va_end(ArgumentPointer);

  // Create normal string
  snprintf(TBuf, sizeof(TBuf), "%s (%s): %s", Status->getName(), StateString[Severity], Tmp);

  // Create string with severity encoding
  snprintf(SBuf, sizeof(SBuf), "%s**", Tmp);
  SBuf[strlen(SBuf)-2] = '\0';
  SBuf[strlen(SBuf)+1] = Severity;

  if (Tmp != ErrorString) free(Tmp);
  
  // Send message to console and log file
  printf("%s\n", TBuf);
  DimClient::sendCommandNB("DColl/Log", TBuf);

  // Update DIM status service (including severity encoding)
  if (Status != NULL) Status->updateService(SBuf, strlen(SBuf)+2);

  // Terminate if message type is fatal
  if (Severity == FATAL) exit(EXIT_FAILURE);
}

// Get configuration data (program terminates if data is missing)
//
// The memory allocated by all calls to this function will be freed by
// the destructor.
char* EvidenceServer::GetConfig(const char *Item, const char *Default) {
  
  // Determine configuration file update time
  DimCurrentInfo ModifyTime("Config/ModifyTime", 0);
  int Time = ModifyTime.getInt(), ItemNo = -1;
  
  // Check if configuration request already in list
  for (unsigned int i=0; i<ConfigList.size(); i++) {
    if (strcmp(ConfigList[i].Name, Item) == 0) {
	  // Return original value if still up to date
	  if (ConfigList[i].Time >= Time) return ConfigList[i].Value;

	  // Otherwise, free memory of old value
	  delete[] ConfigList[i].Name;
	  delete[] ConfigList[i].Value;	  
	  ItemNo = i;
	  break;
	}
  } 
  
  // Make configuration request
  DimRpcInfo Config((char *) "ConfigRequest", (char *) "");
  Config.setData((char *) Item);
  char *Result = Config.getString();
  
  // Terminate if not successful
  if (strlen(Result) == 0) {
    if (Default == NULL) State(FATAL, "Missing configuration data '%s'", Item);
	Result = (char *) Default;
  }

  // Enlarge list if necessary
  if (ItemNo == -1) {
    struct ConfigItem New;
    ConfigList.push_back(New);
	ItemNo = ConfigList.size()-1;
  }

  // Create new entry in item list, allocate memory and copy data to this memory
  ConfigList[ItemNo].Value = new char [strlen(Result)+1];
  ConfigList[ItemNo].Name = new char [strlen(Item)+1];
  strcpy(ConfigList[ItemNo].Name, Item);
  strcpy(ConfigList[ItemNo].Value, Result);
  ConfigList[ItemNo].Time = Time;
    
  // Return address to configuration value  
  return ConfigList[ItemNo].Value;
}


// ====== Static methods ======

// Signal handler (causes pause() and other syscalls to return)
void EvidenceServer::SignalHandler(int) {

  ThisServer->ExitRequest = true;
}

// C++ exception handler
void EvidenceServer::Terminate() {

  string Msg = string(ThisServer->Status->getName()) + ": Caught unhandled exception";

  printf("%s\n", Msg.c_str());
  DimClient::sendCommand("DColl/Log", Msg.c_str());
  
  abort();
}


// Translates DIMInfo to string (memory has to be freed by caller)
// DIM structures are converted to hexadecimal representation
// For string conversion, a terminating \0 is enforced.
char *EvidenceServer::ToString(DimInfo *Item) {

  char *Text;
  int R=0;
  
  if (strlen(Item->getFormat()) != 1) {
    if ((Text = (char *) malloc(3*Item->getSize()+1)) != NULL) {
	  for (int i=0; i<Item->getSize(); i++) sprintf(Text+3*i, "%02x", *((char *) Item->getData() + i)); 
	}
  }
  else {
	switch (*(Item->getFormat())) {
      case 'I':  R = asprintf(&Text, "%d", Item->getInt());   break;
      case 'S':  R = asprintf(&Text, "%hd", Item->getShort());   break;
      case 'F':  R = asprintf(&Text, "%.5f", Item->getFloat());   break;
      case 'D':  R = asprintf(&Text, "%.5f", Item->getDouble());   break;
      case 'X':  R = asprintf(&Text, "%lld", Item->getLonglong());   break;
      case 'C':  *(Item->getString() + Item->getSize() - 1) = '\0'; 
				 R = asprintf(&Text, "%s", Item->getString());
				 break;
      default: return NULL;
	}
  }
  
  return (R == -1) ? NULL : Text;
}

// Checks if service contents indicates not available
bool EvidenceServer::ServiceOK(DimInfo *Item) {

  return !((Item->getSize() == strlen(NO_LINK)+1) &&  
	  (memcmp(Item->getData(), NO_LINK, Item->getSize()) == 0));

}


///////////////////////////
// EvidenceHistory Class //
///////////////////////////

// Organisaztion of history buffer
// F | T S D | T S D | 0 0 ......  | T S D | T S D | 0 -1
//
// F: Offset of oldest entry  T: Time  S: Size  D: Data
// F, T, S: int

// Marker for history buffer
const int EvidenceHistory::WrapMark[] = {0, -1};
const int EvidenceHistory::EndMark[] = {0, 0};

// Constructor
EvidenceHistory::EvidenceHistory() {

  Buffer = NULL;
}

// Destructor
EvidenceHistory::~EvidenceHistory() {

  delete[] Buffer;
}

// Requests service history
bool EvidenceHistory::GetHistory(char *Name) {

  DimCurrentInfo Info((string(Name)+".hist").c_str(), NO_LINK);

  // Check if service available
  if (((Info.getSize() == strlen(NO_LINK)+1) &&  
	  (memcmp(Info.getData(), NO_LINK, Info.getSize()) == 0))) return false;

  delete[] Buffer;
  BufferSize = Info.getSize();
  Buffer = new char [BufferSize];

  memcpy(Buffer, Info.getData(), BufferSize);
  Offset = *(int *) Buffer;
  
  return true;
}

// Returns next item in history buffer
bool EvidenceHistory::Next(int &Time, int &Size, void *&Data) {

  if (Buffer == NULL) return false;

  // Check for wrap around
  if (memcmp(Buffer+Offset, WrapMark, sizeof(WrapMark)) == 0) Offset = 4;
  
  // Check if at end of ring buffer
  if (memcmp(Buffer+Offset, EndMark, sizeof(EndMark)) == 0) return false;

  Time = *(int *) (Buffer + Offset);
  Size = *(int *) (Buffer + Offset + sizeof(int));
  Data = Buffer + Offset + 2*sizeof(int);

  Offset += *((int *) (Buffer + Offset) + 1) + 2*sizeof(int);

  return true;
}
